==> pyautd3-15.0.0-py3-none-linux_armv7l.whl: pyautd3/gain/holo/greedy.py <==
"""
File: greedy.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 28/05/2023
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022-2023 Shun Suzuki. All rights reserved.

"""


import numpy as np
from typing import Iterable, Optional, List
import ctypes

from .constraint import AmplitudeConstraint

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.autd3capi_def import GainPtr
from pyautd3.geometry import Geometry


from pyautd3.gain.gain import IGain


class Greedy(IGain):
    _foci: List[float]
    _amps: List[float]
    _div: Optional[int]
    _constraint: Optional[AmplitudeConstraint]

    def __init__(self):
        self._foci = []
        self._amps = []
        self._div = None
        self._constraint = None

    def add_focus(self, focus: np.ndarray, amp: float) -> "Greedy":
        self._foci.append(focus[0])
        self._foci.append(focus[1])
        self._foci.append(focus[2])
        self._amps.append(amp)
        return self

    def with_phase_div(self, div: int) -> "Greedy":
        self._div = div
        return self

    def with_constraint(self, constraint: AmplitudeConstraint) -> "Greedy":
        self._constraint = constraint
        return self

    def gain_ptr(self, _: Geometry) -> GainPtr:
        size = len(self._amps)
        foci_ = np.ctypeslib.as_ctypes(np.array(self._foci).astype(ctypes.c_double))
        amps = np.ctypeslib.as_ctypes(np.array(self._amps).astype(ctypes.c_double))
        ptr = GainHolo().gain_holo_greedy(foci_, amps, size)
        if self._div is not None:
            ptr = GainHolo().gain_holo_greedy_with_phase_div(ptr, self._div)
        if self._constraint is not None:
            ptr = GainHolo().gain_holo_greedy_with_constraint(
                ptr, self._constraint.ptr()
            )
        return ptr

==> pyautd3-15.3.1-py3-none-linux_armv7l.whl: pyautd3/gain/holo/greedy.py <==
'''
File: greedy.py
Project: holo
Created Date: 21/10/2022
Author: Shun Suzuki
-----
Last Modified: 02/10/2023
Modified By: Shun Suzuki (suzuki@hapis.k.u-tokyo.ac.jp)
-----
Copyright (c) 2022-2023 Shun Suzuki. All rights reserved.

'''


import functools
import numpy as np
from typing import Iterable, Optional, List, Tuple
import ctypes

from .constraint import AmplitudeConstraint

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.autd3capi_def import GainPtr
from pyautd3.geometry import Geometry


from pyautd3.gain.gain import IGain


class Greedy(IGain):
    """Gain to produce multiple foci with greedy algorithm

    - Reference
        * Suzuki, Shun, et al. "Radiation pressure field reconstruction for ultrasound midair haptics by Greedy algorithm with brute-force search,"
          IEEE Transactions on Haptics 14.4 (2021): 914-921.
    """

    _foci: List[float]
    _amps: List[float]
    _div: Optional[int]
    _constraint: Optional[AmplitudeConstraint]

    def __init__(self):
        self._foci = []
        self._amps = []
        self._div = None
        self._constraint = None

    def add_focus(self, focus: np.ndarray, amp: float) -> "Greedy":
        """Add focus

        Arguments:
        - `focus` - Focus point
        - `amp` - Focus amplitude
        """

        assert len(focus) == 3

        self._foci.append(focus[0])
        self._foci.append(focus[1])
        self._foci.append(focus[2])
        self._amps.append(amp)
        return self

    def add_foci_from_iter(
        self, iterable: Iterable[Tuple[np.ndarray, float]]
    ) -> "Greedy":
        """Add foci from iterable

        Arguments:
        - `iterable` - Iterable of focus point and amplitude
        """

        return functools.reduce(
            lambda acc, x: acc.add_focus(x[0], x[1]),
            iterable,
            self,
        )

    def with_phase_div(self, div: int) -> "Greedy":
        self._div = div
        return self

    def with_constraint(self, constraint: AmplitudeConstraint) -> "Greedy":
        """Set amplitude constraint

        Arguments:
        - `constraint` - Amplitude constraint
        """

        self._constraint = constraint
        return self

    def gain_ptr(self, _: Geometry) -> GainPtr:
        size = len(self._amps)
        foci_ = np.ctypeslib.as_ctypes(np.array(self._foci).astype(ctypes.c_double))
        amps = np.ctypeslib.as_ctypes(np.array(self._amps).astype(ctypes.c_double))
        ptr = GainHolo().gain_holo_greedy(foci_, amps, size)
        if self._div is not None:
            ptr = GainHolo().gain_holo_greedy_with_phase_div(ptr, self._div)
        if self._constraint is not None:
            ptr = GainHolo().gain_holo_greedy_with_constraint(
                ptr, self._constraint.ptr()
            )
        return ptr
